#ifndef _testcnfs_h_INCLUDED
#define _testcnfs_h_INCLUDED

#define CNFS \
  CNF (20, false, false) \
  CNF (20, unit1, false) \
  CNF (20, unit2, false) \
  CNF (20, unit3, false) \
  CNF (20, unit4, false) \
  CNF (20, unit5, false) \
  CNF (20, unit6, false) \
  CNF (10, true, false) \
  CNF (10, bin1, false) \
  CNF (10, bin2, false) \
  CNF (10, bin3, false) \
  CNF (20, full2, false) \
  CNF (20, full3, false) \
  CNF (20, full4, false) \
  CNF (10, and1, false) \
  CNF (10, and2, false) \
  CNF (10, and3, false) \
  CNF (10, ite0, false) \
  CNF (10, ite1, false) \
  CNF (10, ite2, false) \
  CNF (10, ite3, false) \
  CNF (10, ite4, false) \
  CNF (10, ite5, false) \
  CNF (10, ite6, false) \
  CNF (10, ite7, false) \
  CNF (10, ite8, false) \
  CNF (10, ite9, false) \
  CNF (10, ite10, false) \
  CNF (10, ite11, false) \
  CNF (10, ite12, false) \
  CNF (10, ite13, false) \
  CNF (10, ite14, false) \
  CNF (10, ite15, false) \
  CNF (10, ite16, false) \
  CNF (10, ite17, false) \
  CNF (10, ite18, false) \
  CNF (10, ite19, false) \
  CNF (10, ite20, false) \
  CNF (10, ite21, false) \
  CNF (10, ite22, false) \
  CNF (10, ite23, false) \
  CNF (10, ite24, false) \
  CNF (10, ite25, false) \
  CNF (10, ite26, false) \
  CNF (10, ite27, false) \
  CNF (10, ite28, false) \
  CNF (10, ite29, false) \
  CNF (10, ite30, false) \
  CNF (10, ite31, false) \
  CNF (10, ite32, false) \
  CNF (10, ite33, false) \
  CNF (10, ite34, false) \
  CNF (10, xor0, false) \
  CNF (10, xor1, false) \
  CNF (10, xor2, false) \
  CNF (10, xor3, false) \
  CNF (10, xor4, false) \
  CNF (20, xor5, false) \
  CNF (10, xor6, false) \
  CNF (20, ph2, false) \
  CNF (20, ph3, false) \
  CNF (20, ph4, false) \
  CNF (20, ph5, false) \
  CNF (20, ph6, false) \
  CNF (10, sqrt3481, false) \
  CNF (10, sqrt3721, false) \
  CNF (10, sqrt2809, false) \
  CNF (10, sqrt12769, false) \
  CNF (10, sqrt5329, false) \
  CNF (10, sqrt6889, false) \
  CNF (10, sqrt4489, false) \
  CNF (10, sqrt5041, false) \
  CNF (10, sqrt6241, false) \
  CNF (10, sqrt7921, false) \
  CNF (10, sqrt16129, false) \
  CNF (10, sqrt11449, false) \
  CNF (10, sqrt10201, false) \
  CNF (10, sqrt10609, false) \
  CNF (10, sqrt9409, false) \
  CNF (10, sqrt11881, false) \
  CNF (10, sqrt63001, false) \
  CNF (10, sqrt259081, false) \
  CNF (10, sqrt1042441, false) \
  CNF (10, prime4, false) \
  CNF (10, prime9, false) \
  CNF (10, prime25, false) \
  CNF (10, prime49, false) \
  CNF (10, prime121, false) \
  CNF (10, prime169, false) \
  CNF (10, prime289, false) \
  CNF (10, prime361, false) \
  CNF (10, prime529, false) \
  CNF (10, prime841, false) \
  CNF (10, prime961, false) \
  CNF (10, prime1369, false) \
  CNF (10, prime1681, false) \
  CNF (10, prime1849, false) \
  CNF (10, prime2209, false) \
  CNF (20, prime65537, false) \
  CNF (20, prime4294967297, true) \
  CNF (20, add4, false) \
  CNF (20, add8, false) \
  CNF (20, add16, false) \
  CNF (20, add32, false) \
  CNF (20, add64, false) \
  CNF (20, add128, false) \
  CNF (20, strash1, false) \
  CNF (20, strash2, false) \
  CNF (20, congr1, false) \
  CNF (20, congr2, false) \
  CNF (20, congr3, false) \
  CNF (10, congr4, false) \
  CNF (20, congr5, false) \
  CNF (20, congr6, false) \
  CNF (10, congr7, false) \
  CNF (10, factor1, false) \
  CNF (10, factor2, false) \
  CNF (10, factor3, false) \
  CNF (10, factor4, false) \
  CNF (10, factor5, false)

#endif
